%% Regular expressions, from the tutorial.
%% At line 150 of nfa_re.erl you will see instructions for making the
%% library buggy. Try it and see what changes!

-module(nfa_re_sig).
-include_lib("eqc/include/eqc.hrl").
-compile(export_all).

var_types() ->
    [{[r, s, t], regex},
     {[x, y, z], character}].

fun_types() ->
    [{nfa_re, char, [character], regex},
     {nfa_re, any, [], regex},
     {nfa_re, empty, [], regex},
     {nfa_re, zero, [], regex},
     {nfa_re, concat, [regex, regex], regex},
     {nfa_re, choice, [regex, regex], regex},
     {nfa_re, star, [regex], regex}].
     %% {nfa_re, plus, [regex], regex}].
     %% {nfa_re, optional, [regex], regex}].

%% Test strings are sequences of as and bs.
character() ->
    elements([$a, $b]).

%% Compare regular expressions by matching them against random
%% strings...
observe(R, regex, Ss) ->
    R1 = nfa_re:compile(R),
    [ nfa_re:run(S, R1) || S <- Ss ].

%% ...and generate 10 random strings at a time.
context() ->
    vector(10, list(character())).

%% Pretty-printing regular expressions.
display(nfa_re, plus, R) ->
    [bracket(R), "+"];
display(nfa_re, star, R) ->
    [bracket(R), "*"];
display(nfa_re, optional, R) ->
    [bracket(R), "?"].
display(nfa_re, any) ->
    ".";
display(nfa_re, zero) ->
    "0";
display(nfa_re, empty) ->
    "()".
display(nfa_re, concat, R1, R2) ->
    [bracket(R1), ";", bracket(R2)];
display(nfa_re, choice, R1, R2) ->
    [bracket(R1), "|", bracket(R2)].

bracket(R) ->
    case atomic(R) of
        true -> laws:display(R);
        false -> ["(", laws:display(R), ")"]
    end.

atomic({call, nfa_re, _, []}) ->
    true;
atomic({call, nfa_re, _, [_]}) ->
    true;
atomic({var, _, _}) ->
    true;
atomic(_) ->
    false.
