%% Testing the Erlang queue module.
%%
%% Try removing the observe() function and you'll see the problem
%% discussed in the slides.

-module(queuesig).
-include_lib("eqc/include/eqc.hrl").
-compile(export_all).

var_types() ->
    [{[x,y,z],int},
     {[q,q2,q3],queue}] ++
        %% Also include some boolean and list operations.
        stdsigs:bool_vars() ++ stdsigs:list_vars(list).

fun_types() ->
    [%% Standard queue operations.
     {queue,new,[],queue},
     {queue,tail,[queue],queue},
     {queue,liat,[queue],queue},
     {queue,lait,[queue],queue},
     {queue,head,[queue],int},
     {queue,daeh,[queue],int},
     {queue,reverse,[queue],queue},
     {queue,in,[int,queue],queue},
     {queue,in_r,[int,queue],queue},
     {queuesig,is_empty,[queue],bool}

     %% A buggy version of is_empty.
     %% {?MODULE,is_empty,[queue],bool},

     %% Testing queues using list as a model.
     %% {?MODULE, init, [list], list},
     %% {lists, last, [list], int},
     %% {queue,to_list,[queue],list}
    ] ++ stdsigs:list_funs(int, list) ++ stdsigs:bool_funs().

init(Xs) ->
    lists:reverse(tl(lists:reverse(Xs))).

%% A buggy version of is_empty.
is_empty({[_|_], _}) ->
    false;
is_empty(_) ->
    true.

%% A buggy version of observe.
%% observe(X, queue) ->
%%     queue:head(X).

%% Compare the contents of a queue instead of its representation.
observe(X, queue) ->
    queue:to_list(X).

%% A generator for queues.
%% queue() ->
%%     ?LET(Xs, list(int()),
%%     queue:from_list(Xs)).

%% Only print laws that have a queue in them somewhere.
interesting(queue) ->
    true;
interesting(_) ->
    false.

%% lait(in(X, Q)) == Q.
%% This property fails without an observe function!
%% Notice that QuickSpec gives us a generator for queues.
prop_in_lait() ->
    ?FORALL(X, int(),
    ?FORALL(Q, laws:symbolic(queuesig, queue),
    queue:lait(queue:in(X, eval(Q))) == eval(Q))).
