%% Erlang's built-in regular expressions, from the tutorial.
%% If you uncomment the 'star' function from fun_types(),
%% the tests won't finish...
-module(re_sig).
-include_lib("eqc/include/eqc.hrl").
-compile(export_all).

var_types() ->
    [{[r, s, t], regex},
     {[x, y, z], character}].

fun_types() ->
    [{?MODULE, char, [character], regex},
     {?MODULE, empty, [], regex},
     {?MODULE, any, [], regex},
     {?MODULE, concat, [regex, regex], regex},
     {?MODULE, choice, [regex, regex], regex}].
     %% {?MODULE, star, [regex], regex}].
     %% {?MODULE, plus, [regex], regex}].
     %% {?MODULE, perhaps, [regex], regex}].

%% Test strings are sequences of as and bs.
character() ->
    elements("ab").

string() ->
    list(character()).

%% Combinators for building regular expressions.
%% Without these the module is untestable!

%% Escape a single character in a regular expression.
char(C) when C >= $a andalso C =< $z ->
    [C];
char(C) when C >= $A andalso C =< $Z ->
    [C];
char(C) when C >= $0 andalso C =< $9 ->
    [C];
char(C) ->
    [$\\, C].

empty() ->
    [].

any() ->
    ".".

concat(R, S) ->
    "(" ++ R ++ ")(" ++ S ++ ")".

choice(R, S) ->
    "(" ++ R ++ ")|(" ++ S ++ ")".

star(R) ->
    "(" ++ R ++ ")*".

plus(R) ->
    "(" ++ R ++ ")+".

perhaps(R) ->
    "(" ++ R ++ ")?".

%% Compare regular expressions by matching them against random
%% strings...
observe(R, regex, Ss) ->
    [
     case re:run(S, "^(" ++ R ++ ")$") of%"
                 {match, _} -> true;
                 nomatch -> false
     end || S <- Ss ].

%% ...and generate 10 random strings at a time.
context() ->
    vector(10,string()).

%% R;() == R.
prop_concatempty() ->
    ?FORALL(R, laws:generate(?MODULE, regex),
    ?FORALL(S, context(),
    observe(R, regex, S) == observe(concat(R, empty()), regex, S))).

%% R|S == S|R.
prop_choicecomm() ->
    ?FORALL([R1, R2], vector(2, laws:generate(?MODULE, regex)),
    ?FORALL(S, context(),
    observe(choice(R1, R2), regex, S) == observe(choice(R2, R1), regex, S))).

%% Evaluating a regular expression should terminate within 500ms.
prop_runs() ->
    ?FORALL(R, laws:generate(?MODULE, regex),
    ?FORALL(Xs, context(),
    ?LET(Result, timeout(500, ?LAZY(observe(R, regex, Xs))),
    Result /= timeout))).

%% Pretty-printing regular expressions.
display(re_sig, plus, R) ->
    [bracket(R), "+"].
display(re_sig, any) ->
    ".";
display(re_sig, zero) ->
    "0";
display(re_sig, empty) ->
    "()".
display(re_sig, concat, R1, R2) ->
    [bracket(R1), ";", bracket(R2)];
display(re_sig, choice, R1, R2) ->
    [bracket(R1), "|", bracket(R2)].

bracket(R) ->
    case atomic(R) of
        true -> laws:display(R);
        false -> ["(", laws:display(R), ")"]
    end.

atomic({call, re_sig, _, []}) ->
    true;
atomic({call, re_sig, _, [_]}) ->
    true;
atomic({var, _, _}) ->
    true;
atomic(_) ->
    false.
