%% The sets module.
%% 
%% The laws generated are a little complicated.
%% This is because add_element is similar to union and we tend to get
%% pairs of laws, one version for union and version for add_element.
%% 
%% QuickSpec produces the following definition
%%   add_element(X,S) := union(S,add_element(X,new()))
%% which states that add_element is a special case of union:
%% to add X to S, you can instead take the union of S
%% with a singleton set containing only X.
%% Try replacing add_element with a singleton function
%% of type elem() -> set(). The laws should get simpler.

-module(setsig).
-include_lib("eqc/include/eqc.hrl").
-compile(export_all).

var_types() ->
    [{[x,y,z],elem},
     {[s,t,u],set},
     {[xs,ys,zs],list}].

fun_types() ->
    [{sets,new,[],set},
     {sets,from_list,[list],set},
     {sets,add_element,[elem,set],set},
     {sets,union,[set,set],set},
     {sets,intersection,[set,set],set}].

elem() ->
    elements([a,b,c,d,e]).

list() ->
    list(elem()).

observe(X, set) ->
    lists:usort(sets:to_list(X)).
